output_var_lines <- function(l,f) {
  .Call("output_var_lines", as.list(l), as.character(f), PACKAGE="mmbgx")
}

merge_transcripts <- function(v1, v2) {
  .Call("merge_transcripts", as.character(v1), as.character(v2), PACKAGE="mmbgx") 
}

makeStructDir <- function(annmapDB, cdfName, geneLevel=FALSE, path=NULL, oligoPkg=NULL) {
  require(annmap)
  require(Rcpp)
  require(inline)
  require(Biostrings)
  require(oligo)
  if(is.null(oligoPkg)) {
    oligoPkg <- paste("pd", gsub("[-_]",".", tolower(cdfName)), sep=".")
  }
  if( !require(package=oligoPkg, character.only=TRUE) ) {
    message("Attempting to download required Bioconductor annotation package: ", oligoPkg)
    source("http://bioconductor.org/biocLite.R")
    biocLite(oligoPkg)
    if(!require(package=oligoPkg, character.only=TRUE) ) {
      stop("Sorry, couldn't load the required Bioconductor annotation package: ", oligoPkg)
    }
  }

  annmapConnect(annmapDB)

  if(cdfName=="HuEx-1_0") {
    stop("Sorry, cdfName \"HuEx-1_0\" is not available, did you mean \"HuEx-1_0-v2\"?")
  }

  arrayType=cdfName
  if(cdfName=="HuEx-1_0-st-v2") {
    arrayType="HuEx-1_0" # annmap doesn't include the -v2 suffix
  }
  if(cdfName=="HG-U133_Plus_2") {
    arrayType="HG-U133Plus2" # annmap doesn't include the underscores
  }

  if(! arrayType %in% allArrays()$array_name) {
    message("You specified cdfName=\"", cdfName, "\", corresponding to array type \"", arrayType, "\", which is not available.\n",
      "The following arrays types are available in database \"", annmapDB, "\": ",
      paste(allArrays()$array_name, collapse=", "))
    annmapDisconnect()
    stop()
  } else {
    message("Setting array type to \"", arrayType, "\"")
    arrayType(arrayType)
  }

  v <- annmapGetParam("version")

  if(!is.null(path)) {
    path=path.expand(path)
  } else {
    if(geneLevel){
      path=path.expand(file.path(.path.package("mmbgx"), "structDirs",paste(cdfName,"gene",sep="-"), v))
    } else {
      path=path.expand(file.path(.path.package("mmbgx"), "structDirs",cdfName, v))
    }
    message(paste("Saving to", path))
  }

  if(file.exists(path)) {
    cat("A structure directory or file already exists at", path, "- overwrite? (y/N) ")
    ans = readline()
    if(tolower(ans) != 'y') {
      stop()
    } else {
      unlink(path, recursive = TRUE)
      dir.create(path, recursive=TRUE)
    }
  } else {
    dir.create(path, recursive=TRUE)
  }

  at = c()
  message(date(), " | Mapping genes to transcripts (~ 1 min)...")
  before = proc.time()[["elapsed"]]
  g2t <- geneToTranscript(allGenes(), as.vector=FALSE)[,c("IN1","stable_id")]
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")
  if(geneLevel) {
    at=allGenes()
    message("Found ", paste(length(at), "genes in annmap."))
  } else {
    at = allTranscripts()
    message("Found ", paste(length(at), "transcripts in annmap"))
  }

  message(date(), " | Mapping ", ifelse(geneLevel, "genes","transcripts")," to probesets and merging (~ 8 mins)...")
  before = proc.time()[["elapsed"]]
  t_ps <- c()
  if(geneLevel) {
    t_ps = transcriptToCdnaprobeset(allTranscripts(), as.vector=FALSE)[,1:2]
    g2t <- data.frame(IN1=g2t$IN1, transcript_id=g2t$stable_id)
    names(t_ps)[1] <- "transcript_id"
    t_ps <- merge(g2t, t_ps, by="transcript_id")[,c("IN1","stable_id")]
    t_ps <- t_ps[order(t_ps$IN1),]
  } else {
    t_ps = transcriptToCdnaprobeset(at, as.vector=FALSE)[,1:2]
  }
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")

  t_ps_merged = data.frame(merge_transcripts(as.character(t_ps[,1]), as.character(t_ps[,2])), stringsAsFactors=FALSE);

  message(date(), " | Mapping probesets to probes (~ 8 mins)...")
  before = proc.time()[["elapsed"]]
  ps_p <- probesetToProbe(as.character(unique(t_ps_merged$stable_id)),as.vector=FALSE)[,1:2];
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")

  message(date(), " | Merging ", ifelse(geneLevel, "genes","transcripts"), ", probesets and probes (~ 1 min)...")
  before = proc.time()[["elapsed"]]
  colnames(ps_p)[1] <- "stable_id"
  map <- merge(t_ps_merged, ps_p, by="stable_id", all.x=TRUE, sort=FALSE)
  map <- map[order(map$IN1),]
  map <- cbind(map, t_index=match(map$IN1, unique(map$IN1)))
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")

  data(pmSequence, package=oligoPkg)

  message(date(), " | Merging ", ifelse(geneLevel, "genes","transcripts"),", probesets and probes with probe indexes (~ 6 min)...")
  before = proc.time()[["elapsed"]]
  oligomap <- data.frame(a_index=(1:nrow(pmSequence)), sequence=as.character(complement(pmSequence$sequence)))
  map2 <- merge(map, oligomap, by="sequence", sort=FALSE)
  map2 <- map2[order(map2$IN1),]
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")

  message(date(), " | Outputting structure files (~ 4 mins)...")
  before = proc.time()[["elapsed"]]
  write(unique(as.character(map2$IN1)), ncolumns=1,file=file.path(path, "PSids.txt"))

  spl <- split(map2$a_index, map2$sequence)
  o <- output_var_lines(spl, file.path(path, "PMids.txt"))

  map2 <- cbind(map2, p_index=match(map2$sequence, names(spl)))

  spl <- split(map2$t_index, map2$sequence)
  spl <- lapply(spl, function(x) { return(x-1) })
  o <- output_var_lines(spl, file.path(path, "PM2PS.txt"))

  write(letterFrequency(BStringSet(names(spl)), letters = "CG"), ncolumns=1, file=file.path(path, "PMcats.txt"))

  spl <- split(map2$p_index, map2$t_index)
  spl <- lapply(spl, function(x) {return(x-1) })
  o <- output_var_lines(spl, file.path(path, "PS2PM.txt") )

#  cat("", file=file.path(path, "PS2PS.txt")) # not bothering with this

  conn <- db(get(oligoPkg))
  sql <- paste("SELECT type FROM type_dict", "WHERE type_id=\"control->bgp->antigenomic\"")
  bgCode <- as.integer(dbGetQuery(conn, sql)[[1]])
  if (length(bgCode) != 1) 
    stop("Can't find proper id for control-bgp-antigenomic probes")
  sql <- paste("SELECT fid FROM pmfeature", "INNER JOIN featureSet USING(fsetid)", 
    "WHERE", paste("type=", bgCode, sep = ""))
  tmp <- dbGetQuery(conn, sql)
  idx <- match(tmp$fid, pmSequence$fid)
  write( (1:nrow(pmSequence))[idx] , ncolumns=1, file=file.path(path, "BGids.txt") )
  write( oligo:::gcCounts(pmSequence$sequence[idx]), ncolumns=1, file=file.path(path, "BGcats.txt") )


  if(!geneLevel) {
    write(allGenes()$stable_id, ncolumns=1, file=file.path(path, "geneIDs.txt"))

    spl <- split(g2t$stable_id, g2t$IN1)
    zz <- file(file.path(path, "geneTranscripts.txt") , "w") 
    for(i in 1:length(spl)) {
      cat(paste(spl[[i]]), '\n', file=zz)
    }
    close(zz)
  }

  annmapDisconnect()
  cat("done (", (proc.time()[["elapsed"]] - before)/60.0, " mins).\n", sep="")

}

