////////////////////////////////////////////////////////////////
//
// Copyright (C) 2005 Affymetrix, Inc.
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License (version 2) as 
// published by the Free Software Foundation.
// 
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License 
// along with this program;if not, write to the 
// 
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
////////////////////////////////////////////////////////////////
/**
 * @file   ChpCheck.h
 * @author Chuck Sugnet
 * @date   Tue Apr 25 18:03:02 2006
 * 
 * @brief  Class for doing a check of two CHP files after regression run.
 */

#ifndef CALVINCHPCHECK_H
#define CALVINCHPCHECK_H

#include <string>
#include <set>
#include <stdlib.h>
#include <sys/stat.h>
#include <math.h>
#include <algorithm>
#include <vector>
#include "Util.h"
#include "Verbose.h"
#include "RegressionCheck.h"
#include "CHPMultiDataData.h"
#include "CHPMultiDataFileReader.h"
#include "StringUtils.h"
#include "portability/affy-base-types.h"

using namespace affymetrix_calvin_io;
using namespace affymetrix_calvin_data;
using namespace std;

/**
 * Class for testing that CHP files are the same +/- some epsilon. Also checks
 * to make sure that at least some of the headers are the same
 * (times/dates/versions change so not checked). 
 */
class CalvinChpCheck : public RegressionCheck {

public:

  /** 
   * Constructor.
   * @param generated - vector of filenames to be tested.
   * @param gold - matching vector of filenames for gold standard data to test against.
   * @param eps - epsilon: small amount that two floats can differ by,
   * but still be considered equivalent.
   */
  CalvinChpCheck(std::vector<std::string> &generated, std::vector<std::string> &gold, double eps, 
                 int diffAllowed, bool genotypeOnly, const std::string &prefix) {
    if(generated.size() != gold.size()) {
      Err::errAbort("CalvinChpCheck::CalvinChpCheck() - generated and gold vectors must be same size.");
    }
    m_Generated = generated;
    m_Gold = gold;
    m_Eps = eps;
    m_DiffAllowed = diffAllowed;
    m_GenotypeOnly = genotypeOnly;
    m_MaxErrorsReport = 30;
    m_CurrentErrorsReported = 0;
    m_Prefix = prefix;
  }

  void reportError(const string &err) {
    if(m_CurrentErrorsReported++ < m_MaxErrorsReport) {
      Verbose::out(1, "Error encountered: " + err);
      if(m_CurrentErrorsReported >= m_MaxErrorsReport) {
        Verbose::out(1, "Maximum number of errors reported.");
      }
    }
  }

  /** 
   * Check to make sure that two files are the same +/- some epsilon. 
   * @param msg - Fills in an error message if test fails, empty string otherwise.
   * @return - Returns true if files are close enough, false otherwise.
   */
  bool check(std::string &msg) {
    bool success = true;
    for(unsigned int i = 0; i < m_Generated.size(); i++) {
      m_Generated[i] = Util::getPathName(m_Generated[i].c_str());
      m_Gold[i] = Util::getPathName(m_Gold[i].c_str());
      CHPMultiDataData generatedChp, goldChp;
      CHPMultiDataFileReader chpReader;
      chpReader.SetFilename(m_Generated[i]);
      chpReader.Read(generatedChp);
      chpReader.SetFilename(m_Gold[i]);
      chpReader.Read(goldChp);
      if(!headersSame(goldChp, generatedChp, msg)) {
        success = false;
      }
      if(!genotypeDataSame(goldChp, generatedChp, msg, GenotypeMultiDataType)) {
        success = false;
      }
      if(!genotypeDataSame(goldChp, generatedChp, msg, GenotypeControlMultiDataType)) {
        success = false;
      }
      if(!m_GenotypeOnly && !expressionDataSame(goldChp, generatedChp, msg, ExpressionMultiDataType)) {
        success = false;
      }
      if(!m_GenotypeOnly && !expressionDataSame(goldChp, generatedChp, msg, ExpressionControlMultiDataType)) {
        success = false;
      }
    }
    return success;
  }

private:

  /** 
   * Utility function to see if a condition is true and if not to add
   * the message to our summary of messages.
   * @param condition - boolean to be tested.
   * @param msg - informative message for this condition (if fails)
   * @param summary - ongoing collection of messages.
   * @return - condition test results.
   */
  bool checkMsg(bool condition, const std::string &msg, std::string &summary) {
    if(!condition) {
      summary += msg;
      reportError(msg);
    }
    return condition;
  }

  /** 
   * Check to see if two generic types are equivalent. If floats then
   * makes sure they are withing epsilon (small number). Otherwise
   * tested as strings which should work for other types.
   * @param gold - Type with "correct" values.
   * @param generated - Type with values to be tested.
   * @param eps - epsilon, two floats are called same if they are within this amount.
   * @return - true if same, false otherwise.
   */
  bool paramNamValTypeSame(ParameterNameValueType &gold, ParameterNameValueType &generated, double eps) {
    string goldName, genName, goldValue, genValue;
    /* For ease in debugger... */
    goldName = StringUtils::ConvertWCSToMBS(gold.GetName());
    genName = StringUtils::ConvertWCSToMBS(generated.GetName());
    goldValue = StringUtils::ConvertWCSToMBS(gold.ToString());
    genValue = StringUtils::ConvertWCSToMBS(generated.ToString());
    //    bool match = gold.GetName() == generated.GetName();
    bool match = true;
    if(gold.GetParameterType() == ParameterNameValueType::FloatType) {
      float goldFloat = gold.GetValueFloat();
      float genFloat = generated.GetValueFloat();
      double maxDiff = 0;
      bool ok = checkFloat(goldFloat, genFloat, eps, match, maxDiff);
      match &= ok;
      if(!ok && m_CurrentErrorsReported < m_MaxErrorsReport) {
        printf("Difference found. gold: %.20f test %.20f\n", goldFloat, genFloat);
      }
    }
    // othewise compare as string.
    else {
      match &= gold.ToString() == generated.ToString();
    }
    return match;
  }

  /** 
   * Utility function for checking if two vectors of generic types are the same.
   * @param gold - Vector with the "correct" values.
   * @param generated - Vector to be tested against the gold values.
   * @param eps - epsilon, two floats are called same if they are within this amount.
   * @return - true if same, false othewise.
   */
  bool paramVectorsSame(vector<ParameterNameValueType> &gold, vector<ParameterNameValueType> &generated, double eps) {
    bool match = true;
    if(gold.size() != generated.size())
      return false;
    for(int i = 0; i < gold.size(); i++) {
      match &= paramNamValTypeSame(gold[i], generated[i], eps);
    }
    return match;
  }
  
  static void fillInToIgnore(std::set<std::wstring> &ignoreMap, const std::wstring &prefix) {
    ignoreMap.clear();
    ignoreMap.insert(L"analysis-guid");
    ignoreMap.insert(L"cel-files");
    ignoreMap.insert(L"chrX-file");
    ignoreMap.insert(L"cvs-id"); 
    ignoreMap.insert(L"exec-guid");
    ignoreMap.insert(L"free-mem");    
    ignoreMap.insert(L"program-id"); 
    ignoreMap.insert(L"program-version");
    ignoreMap.insert(L"system-arch");
    ignoreMap.insert(L"system-hostname");
    ignoreMap.insert(L"system-meminfo");
    ignoreMap.insert(L"system-username");
    ignoreMap.insert(L"time-str"); 
    ignoreMap.insert(prefix + L"-analysis-guid");
    ignoreMap.insert(prefix + L"-block-size");
    ignoreMap.insert(prefix + L"-cvs-id");
    ignoreMap.insert(prefix + L"-exec-guid");
    ignoreMap.insert(prefix + L"-time-str");
    ignoreMap.insert(prefix + L"-version");
  }
  
  bool ParameterNameValueTypeMostlySame(ParameterNameValueTypeList &gold, ParameterNameValueTypeList &test, 
                                        std::string &msg, std::wstring &prefix) {
    bool same = true;
    map<wstring,ParameterNameValueType> testMap;
    map<wstring,ParameterNameValueType>::iterator testMapIter;
    ParameterNameValueTypeList::iterator goldIter; 
    ParameterNameValueTypeList::iterator testIter;
    std::set<std::wstring> toIgnore;
    // Load up test as a map which will be queried by items in gold.
    for(testIter = test.begin(); testIter != test.end(); ++testIter) {
      testMap[testIter->GetName()] = *testIter;
    }
    fillInToIgnore(toIgnore, prefix); // things know to change like user, time, etc.
    for(goldIter = gold.begin(); goldIter != gold.end(); ++goldIter) {
      // ignore items that are in the ignoreMap
      if(toIgnore.find(goldIter->GetName()) == toIgnore.end()) {
        testMapIter = testMap.find(goldIter->GetName());
        if(testMapIter == testMap.end()) {
          msg += " Error: Test missing field: '" + StringUtils::ConvertWCSToMBS(goldIter->GetName()) + "'";
          same = false;
        }
        else {
          string goldName, genName, goldValue, genValue;
          /* For ease in debugger... */
          goldName = StringUtils::ConvertWCSToMBS(goldIter->GetName());
          genName = StringUtils::ConvertWCSToMBS(testMapIter->second.GetName());
          goldValue = StringUtils::ConvertWCSToMBS(goldIter->ToString());
          genValue = StringUtils::ConvertWCSToMBS(testMapIter->second.ToString());
          if(!paramNamValTypeSame(testMapIter->second, *goldIter, m_Eps) &&
             Util::fileRoot(genValue) != Util::fileRoot(goldValue)) {
            msg += " Error: for field '" + goldName + "' expecting: '" + goldValue + 
              "' got: '" + genValue + "'";
            same = false;
          }
        }
      }
    }
    return same;
  }

  /** 
   * Are the headers for to CHPMultiDataData CHP files the same?
   * Ignore some fields known to change like time, run id, etc.
   * @param goldChp - "correct" chp file.
   * @param generatedChp - chp file to be tested against gold standard.
   * @param msgs - ongoing list of messages about what when wrong if anything.
   * @return - true if headers are equivalent, false otherwise.
   */
  bool headersSame(CHPMultiDataData &goldChp, CHPMultiDataData &generatedChp, std::string &msgs) {
    ///@todo Re-enable CHP header check once have new regression data
    return true;
    //
    bool success = true;
    FileHeader *goldHeader = NULL, *genHeader = NULL;
    goldHeader = goldChp.GetFileHeader();
    genHeader = generatedChp.GetFileHeader();
    /* Check the version. */
    success &= checkMsg(goldHeader->GetVersion() == genHeader->GetVersion(), "Error: different chp versions.", msgs);

    /* Check the group numbers. */
    success &= checkMsg(goldHeader->GetNumDataGroups() == genHeader->GetNumDataGroups(), 
                        "Error: Different group counts.", msgs);

    /* What type of array chp is from. */
    success &= checkMsg(goldChp.GetArrayType() == generatedChp.GetArrayType(), 
                        "Error: Different array types.", msgs);
    
    /* Algorithm used. */
    success &= checkMsg(goldChp.GetAlgName() == generatedChp.GetAlgName(), 
                        "Error: Different algorithm names.", msgs);
    
    /* Algorithm version. */
    success &= checkMsg(goldChp.GetAlgVersion() == generatedChp.GetAlgVersion(), 
                        "Error: Different algorithm names.", msgs);
    
    
    ParameterNameValueTypeList goldList = goldChp.GetSummaryParams();
    ParameterNameValueTypeList genList = generatedChp.GetSummaryParams();
    ParameterNameValueTypeList::iterator genIter, goldIter;
    std::wstring prefix = StringUtils::ConvertMBSToWCS(m_Prefix);
    success &= ParameterNameValueTypeMostlySame(goldList, genList, msgs, prefix);
    goldList = goldChp.GetAlgParams();
    genList = generatedChp.GetAlgParams();
    success &= ParameterNameValueTypeMostlySame(goldList, genList, msgs, prefix);
//     success &= checkMsg(goldList.size() == genList.size(), 
//                         "Error: Different number of parameters.", msgs);
//     for(genIter = genList.begin(), goldIter = goldList.begin();
//         genIter != genList.end() && goldIter != goldList.end();
//         ++genIter, ++goldIter) {
//       string goldName, genName, goldValue, genValue;
//       /* For ease in debugger... */
//       goldName = StringUtils::ConvertWCSToMBS(goldIter->GetName());
//       genName = StringUtils::ConvertWCSToMBS(genIter->GetName());
//       goldValue = StringUtils::ConvertWCSToMBS(goldIter->ToString());
//       genValue = StringUtils::ConvertWCSToMBS(genIter->ToString());
//       // skip the parameters that should be changing
//       if(goldIter->GetName() == L"analysis-guid" || goldIter->GetName() == L"exec-guid" ||
//          goldIter->GetName() == L"cvs-id" || goldIter->GetName() == L"program-id" || 
//          goldIter->GetName() == L"chrX-file" || goldIter->GetName() == L"time-str" || 
//          goldIter->GetName() == L"cel-files") {
//         continue;
//       }
//       bool match = (goldIter->GetName() == genIter->GetName());
//       // if float check that we're close enough
//       if(goldIter->GetParameterType() == ParameterNameValueType::FloatType) {
//         float goldFloat = goldIter->GetValueFloat();
//         float genFloat = genIter->GetValueFloat();
//         double maxDiff = 0;
//         match &= checkFloat(goldFloat, genFloat, m_Eps, match, maxDiff);
//       }
//       // othewise compare as string.
//       else {
//         string goldString, genString;
//         goldString = StringUtils::ConvertWCSToMBS(goldIter->ToString());
//         genString = StringUtils::ConvertWCSToMBS(genIter->ToString());
//         if(goldIter->ToString() != genIter->ToString()) {
//           match &= false;
//           reportError("String values different. gold: '" + goldString + "' test: '" + genString + "'");
//         }
//       }
//       success &= checkMsg(match, "Error: values not equivalent for: '" + StringUtils::ConvertWCSToMBS(goldIter->GetName()) + "' '" + 
//                           StringUtils::ConvertWCSToMBS(goldIter->ToString()) + "' vs. '" + 
//                           StringUtils::ConvertWCSToMBS(genIter->ToString()) + "'", msgs);
//    }
    return success;
  }


  /** 
   * Are two floats within epsilon (small value) of eachother?
   * 
   * @param gold - correct value.
   * @param gen - value to be tested.
   * @param eps - epsilon, small number they can be different but still equivalent.
   * @param success - boolean set to true false if floats not equivalent.
   * @param maxDiff - Maximum difference seen so far.
   * @return - true if floats equivalent, false otherwise.
   */
  bool checkFloat(float gold, float gen, double eps, bool &success, double &maxDiff) {
    double diff = fabs(gold - gen);
    if(!Util::isFinite(gold) || !Util::isFinite(gen)) {
      Verbose::out(1, "Non-finite floating point values encountered.");
      success = false;
      reportError("Non-finite floats. gold: '" + ToStr(gold) + "' test: '" + ToStr(gen) + "'");
    }
    maxDiff = Max(diff, maxDiff);
    if(diff > eps) {
      success = false;
      reportError("Different floats. gold: '" + ToStr(gold) + "' test: '" + ToStr(gen) + "'");
    }
    return success;
  }
  
  /** 
   * Determine if two gentoyping chp file entries are equivalent.
   * @param goldEntry - "correct" entry.
   * @param generatedEntry - entry to be tested against gold standard.
   * @param localSuccess - set to to false if entries not equivalent.
   * @param maxDiff - maximum difference seen so far.
   * @param numDiff - number of differences seen so far, incremented if these entries not equivalent.
   */
  void checkGenotypeEntry(ProbeSetMultiDataGenotypeData &goldEntry, ProbeSetMultiDataGenotypeData &generatedEntry, 
                         bool &localSuccess, double &maxDiff, int &numDiff) {
    if(goldEntry.name != generatedEntry.name) {
      localSuccess = false;
      reportError("Different names. gold: '" + goldEntry.name + "' test: '" + generatedEntry.name);
    }
    if(goldEntry.call != generatedEntry.call) {
      localSuccess = false;
      reportError("Different calls for snp: '" + goldEntry.name + "'. gold: '" + 
                  goldEntry.name + "' test: '" + generatedEntry.name + "'");
    }
    checkFloat(goldEntry.confidence, generatedEntry.confidence, m_Eps, localSuccess, maxDiff);
    if(!paramVectorsSame(goldEntry.metrics, generatedEntry.metrics, m_Eps)) {
      localSuccess = false;
      reportError("Param vectors different for snp: '" + goldEntry.name + "'.");
    }
    if(!localSuccess) {
      numDiff++;
    }
  }

  /** 
   * Determine if two expression chp file entries are equivalent.
   * @param goldEntry - "correct" entry.
   * @param generatedEntry - entry to be tested against gold standard.
   * @param localSuccess - set to to false if entries not equivalent.
   * @param maxDiff - maximum difference seen so far.
   * @param numDiff - number of differences seen so far, incremented if these entries not equivalent.
   */
  void checkExpressionEntry(ProbeSetMultiDataExpressionData &goldEntry, ProbeSetMultiDataExpressionData &generatedEntry, 
                         bool &localSuccess, double &maxDiff, int &numDiff) {
    if(goldEntry.name != generatedEntry.name) {
      localSuccess = false;
      reportError("Different names. gold: '" + goldEntry.name + "' test: '" + generatedEntry.name + "'");
    }
    checkFloat(goldEntry.quantification, generatedEntry.quantification, m_Eps, localSuccess, maxDiff);
    if(!paramVectorsSame(goldEntry.metrics, generatedEntry.metrics, m_Eps)) {
      localSuccess = false;
      reportError("Param vectors different for probeset: '" + goldEntry.name + "'.");
    }
    if(!localSuccess) {
      numDiff++;
    }
  }

  /** 
   * Check to see if the genotype entries for MultiDataType are the same.
   * @param goldChp - "correct" data.
   * @param generatedChp - chp file data to test against gold standard.
   * @param msgs - any ongoing messages for user.
   * @param dataType - Either genotype or control genotype, they are conceptually  separate in chp file.
   * @return - true if genotype entries are equivalent, false otherwise.
   */
  bool genotypeDataSame(CHPMultiDataData &goldChp, CHPMultiDataData &generatedChp, std::string &msgs, MultiDataType dataType) {
    bool success = true;
    double maxDiff = -1;
    int numDiff = 0;
    int numGenotyping = goldChp.GetEntryCount(dataType);
    int genNum = generatedChp.GetEntryCount(dataType);
    assert(dataType == GenotypeMultiDataType || dataType == GenotypeControlMultiDataType);
    Verbose::out(4, ToStr(numGenotyping) + " " + StringUtils::ConvertWCSToMBS(MultiDataDataSetNames[dataType]) + " genotyping probesets.");
    if(!checkMsg(numGenotyping == genNum, "Wrong number of genotyping probesets.", msgs))
      return false;
    // if nothing to check, skip it.
    if(numGenotyping == 0)
      return true;
    for(int i = 0; i < numGenotyping; i++) {
      bool localSuccess = true;
      ProbeSetMultiDataGenotypeData goldEntry, generatedEntry;
      goldChp.GetGenotypeEntry(dataType, i, goldEntry);
      generatedChp.GetGenotypeEntry(dataType, i, generatedEntry);
      checkGenotypeEntry(goldEntry, generatedEntry, localSuccess, maxDiff, numDiff);
      success &= localSuccess;
    }
    if(maxDiff > m_Eps) {
      Verbose::out(1, "Max diff: " + ToStr(maxDiff) + " is greater than expected (" + ToStr(m_Eps) + ")");
      Verbose::out(1, ToStr(numDiff) + " of " + ToStr(numGenotyping) + " (" + 
                   ToStr(100.0 * numDiff/numGenotyping) + "%) were different.");
    }
    if(!success && m_DiffAllowed >= numDiff) 
        success = true;
    if(!success) {
      msgs += "Error: " + generatedChp.GetFilename() + " is different from " + goldChp.GetFilename() + ". ";
    }
    std::string res = "different";
    if(success) 
        res = "equivalent";
    Verbose::out(1, generatedChp.GetFilename() + ToStr(" chip is ") + res + " max diff is: " + ToStr(maxDiff));
    return success;
  }
  
  /** 
   * Check to see if the expression entries for MultiDataType are the same.
   * @param goldChp - "correct" data.
   * @param generatedChp - chp file data to test against gold standard.
   * @param msgs - any ongoing messages for user.
   * @param dataType - Either expression or control expression, they are conceptually separate in chp file.
   * @return - true if expression entries are equivalent, false otherwise.
   */
  bool expressionDataSame(CHPMultiDataData &goldChp, CHPMultiDataData &generatedChp, std::string &msgs, MultiDataType dataType) {
    bool success = true;
    double maxDiff = -1;
    int numDiff = 0;
    assert(dataType == ExpressionMultiDataType || dataType == ExpressionControlMultiDataType);
    int numExpr = goldChp.GetEntryCount(dataType);
    int genNum = generatedChp.GetEntryCount(dataType);
    Verbose::out(4, ToStr(numExpr) + " " + StringUtils::ConvertWCSToMBS(MultiDataDataSetNames[dataType]) + " expression probesets.");
    if(!checkMsg(numExpr == genNum, "Wrong number of expression probesets.", msgs))
      return false;
    // if nothing to check, skip it.
    if(numExpr == 0)
      return true;
    for(int i = 0; i < numExpr; i++) {
      bool localSuccess = true;
      ProbeSetMultiDataExpressionData goldEntry, generatedEntry;
      goldChp.GetExpressionEntry(dataType, i, goldEntry);
      generatedChp.GetExpressionEntry(dataType, i, generatedEntry);
      checkExpressionEntry(goldEntry, generatedEntry, localSuccess, maxDiff, numDiff);
      success &= localSuccess;
    }
    if(maxDiff > m_Eps) {
      Verbose::out(1, "Max diff: " + ToStr(maxDiff) + " is greater than expected (" + ToStr(m_Eps) + ")");
      Verbose::out(1, ToStr(numDiff) + " of " + ToStr(numExpr) + " (" + 
                   ToStr(100.0 * numDiff/numExpr) + "%) were different.");
    }
    if(!success) {
      msgs += "Error: " + generatedChp.GetFilename() + " is different from " + goldChp.GetFilename() + ". ";
      
    }
    std::string res = "different";
    if(success) 
        res = "equivalent";
    Verbose::out(1, generatedChp.GetFilename() + ToStr(" chip is ") + res + " max diff is: " + ToStr(maxDiff));
    return success;
  }

  /// Filenames for the gold standard or correct data chp files.
  std::vector<std::string> m_Gold;
  /// Matching filenames for the chp files to be tested.
  std::vector<std::string> m_Generated;
  /// Epsilon, small value that two floats can differ by but still be considered equivalent.
  double m_Eps;
  /// Are we only testing the genotyping entries?
  bool m_GenotypeOnly;
  /// How many differences will we tolerate?
  int m_DiffAllowed;
  /// How many errors to print before being quiet.
  int m_MaxErrorsReport;
  /// How many errors have we printed so far?
  int m_CurrentErrorsReported;
  /// What is the expected prefix for parameter names?
  std::string m_Prefix;
  
};

#endif /* CALVINCHPCHECK_H */
