/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *                2005-2008  Ernest Turro <ernest.turro@ic.ac.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <vector>
#include <cstdlib>
#include <dirent.h>
#include "bgx.hh"

using namespace std;

vector<int> * getIntsByLine(const char *file, int lines);

// First argument: path to infile.txt (must exist)
// Optional second argument: path to output run directory (must exist)
#ifdef USING_R // Avoid clash between main here and main from R command
int main2(int argc, const char* argv[])
#else
int main(int argc, const char* argv[])
#endif
{
  int samples,conditions,probes,genes,numberOfCategories,seed,iter,burnin,
    *samplesets, *categories, 
    subsample, adaptive, batch_size;
  double *gamma, *delta, s_jmp, mu_jmp, sigma_jmp, *pm, optimalAR;
  char *dirname, *basepath;

  if(argc<2){
    cerr << "No input file specified" << endl;
    exit(1);
  }
  ifstream infile(argv[1]);
  if(!infile){
    cerr << "Cannot open input file" << endl;
    exit(2);
  }

  // Find input base path
  string inputpath(argv[1]);
  int i = inputpath.size()-1;
  while(inputpath[i] != '/' && i>=0){
    inputpath.erase(i,1);
    i--;
  }

  // Initialise base path for output runs
  basepath = new char[1024];
  if(argc>2) {
    if(opendir(argv[2]) == NULL) {
      cerr << "Output path does not exist" << endl;
      exit(3);
    } else {
      strcpy(basepath,argv[2]); // We create the run directories in specified directory
    }
  } else {
    strcpy(basepath,"."); // We create the run directories in current directory
  }
  
  samples = conditions = probes = genes = seed = iter = burnin = numberOfCategories = adaptive = batch_size = -1;
  s_jmp = mu_jmp = sigma_jmp= optimalAR = -1.0;
  
  string samplesets_file="";
  string ps2pm_file="";
  string pm2ps_file="";
  string gamma_file="";
  string delta_file="";

  string categories_file="";
  string PM_file="";
  string genesToWatch_file=""; 
  string firstProbeInEachGeneToWatch_file="";
  char sampleNames_[1024]; // will hold the names of the CEL files as one C string
  string filesToCopyFile= "";

  string temp_str;
  while(infile >> temp_str){
    if(temp_str=="samples") infile >> samples;
    else if(temp_str=="conditions") infile >> conditions;
    else if(temp_str=="probes") infile >> probes;
    else if(temp_str=="probesets") infile >> genes;
    else if(temp_str=="numberOfCategories") infile >> numberOfCategories;

    else if(temp_str=="PM2PS") infile >> pm2ps_file;
    else if(temp_str=="PS2PM") infile >> ps2pm_file;
    else if(temp_str=="gamma") infile >> gamma_file;
    else if(temp_str=="delta2") infile >> delta_file;

    else if(temp_str=="SampleSets") infile >> samplesets_file;
    else if(temp_str=="Categories") infile >> categories_file;
    else if(temp_str=="PM") infile >> PM_file;
    else if(temp_str=="seed") infile >> seed;
    else if(temp_str=="sweeps") infile >> iter;
    else if(temp_str=="burn-in") infile >> burnin;
    else if(temp_str=="S_jmp") infile >> s_jmp;
    else if(temp_str=="Mu_jmp") infile >> mu_jmp;
    else if(temp_str=="Sigma_jmp") infile >> sigma_jmp;
    else if(temp_str=="Adaptive") infile >> adaptive;
    else if(temp_str=="BatchSize") infile >> batch_size;
    else if(temp_str=="OptimalAR") infile >> optimalAR;
    else if(temp_str=="CEL_file") infile.getline(sampleNames_, 1024);
    else if(temp_str=="filesToCopy") infile >> filesToCopyFile;
    else if(temp_str[0]=='#') getline(infile,temp_str);
    else{
        cerr << "Malformed input file at unrecognised token " << temp_str << endl;
	      exit(3);
    }
  }

  if(samples==-1){
    cerr << "Number of samples not specified in input file" << endl;
    exit(11);
  }
  if(conditions==-1){
    cerr << "Number of conditions not specified in input file" << endl;
    exit(11);
  }
  if(probes==-1){
    cerr << "Number of probes not specified in input file" << endl;
    exit(11);
  }
  if(genes==-1){
    cerr << "Number of probesets not specified in input file" << endl;
    exit(11);
  }
  if(numberOfCategories==-1){
    cerr << "Number of categories not specified in input file" << endl;
    exit(11);
  }
  if(pm2ps_file==""){
    cerr << "PM2PS file not specified in input file" << endl;
    exit(11);
  }
  if(ps2pm_file==""){
    cerr << "PS2PM file not specified in input file" << endl;
    exit(11);
  }
  if(gamma_file==""){
    cerr << "Gamma file not specified in input file" << endl;
    exit(11);
  }
  if(delta_file==""){
    cerr << "Delta file not specified in input file" << endl;
    exit(11);
  }
  if(samplesets_file==""){
    cerr << "SampleSets file not specified in input file" << endl;
    exit(11);
  }
  if(categories_file==""){
    cerr << "categories file not specified in input file" << endl;
    exit(11);
  }
  if(PM_file==""){
    cerr << "PM file not specified in input file" << endl;
    exit(11);
  }
  if(seed==-1){
    cerr << "Random number seed not specified in input file" << endl;
    exit(11);
  }
  if(iter==-1){
    cerr << "Number of sampling sweeps not specified in input file" << endl;
    exit(11);
  }
  if(burnin==-1){
    cerr << "Number of burn-in sweeps not specified in input file" << endl;
    exit(11);
  }
  if(s_jmp==-1){
    cerr << "Spread of jumps for RWM on S not specified in input file" << endl;
    exit(11);
  }
  if(mu_jmp==-1){
    cerr << "Spread of jumps for RWM on Mu not specified in input file" << endl;
    exit(11);
  }
  if(sigma_jmp==-1){
    cerr << "Spread of jumps for RWM on Sigma not specified in input file" << endl;
    exit(11);
  }

  if(adaptive==-1){
    cerr << "You have not specified whether to use adaptive MCMC" << endl;
    exit(11);
  }
  if(adaptive){
    if(batch_size==-1 || optimalAR==-1){
      cerr << "You have not specified a batch size and/or optimal acceptance ratio" << endl;
      exit(11);
    }
  }


  vector<string> filesToCopy;
  if(filesToCopyFile != "") {
    ifstream fil((inputpath + filesToCopyFile).c_str());
    string temp;
    while(1) {
      if(fil.eof()) break;
      fil >> temp;
      filesToCopy.push_back(inputpath + temp);
    }
    fil.close();
  }


  vector<char *> sampleNames;
  char *pch = strtok(sampleNames_, " ,");
  while(pch != NULL){
    sampleNames.push_back(pch);
    pch = strtok(NULL, " ,");
  }

  ifstream SampleSets((inputpath + samplesets_file).c_str());
  if(SampleSets==NULL){
    cerr << "Could not open the Sample Sets file" << endl;
    exit(4);
  }
  samplesets = new int[conditions];
  for(int i=0; i<conditions; ++i)
    SampleSets >> samplesets[i];
  SampleSets.close();

  ifstream Categories((inputpath + categories_file).c_str());
  if(Categories==NULL){
    cerr << "Could not open the Categories file" << endl;
    exit(5);
  }
  categories = new int[probes];
  for(int i=0; i < probes; i++)
    Categories >> categories[i];
  Categories.close();

  ifstream PM((inputpath + PM_file).c_str());
  if(PM==NULL){
    cerr << "Could not open the PM file" << endl;
    exit(6);
  }
  pm = new double[probes*samples];
  for(int i=0; i<probes*samples; ++i){
    PM >> pm[i];
  }
  PM.close();

  vector<int> *pm2ps = getIntsByLine((inputpath + pm2ps_file).c_str(), probes);
  vector<int> *ps2pm = getIntsByLine((inputpath + ps2pm_file).c_str(), genes);

  ifstream gamma_((inputpath + gamma_file).c_str());
  if(gamma_==NULL){
    cerr << "Could not open the gamma file" << endl;
    exit(5);
  }
  gamma= new double[numberOfCategories*samples];
  for(int i=0; i<numberOfCategories*samples; ++i) {
    gamma_ >> gamma[i];
  }
  ifstream delta_((inputpath + delta_file).c_str());
  if(delta_==NULL){
    cerr << "Could not open the delta file" << endl;
    exit(5);
  }
  delta= new double[numberOfCategories*samples];
  for(int i=0; i<numberOfCategories*samples; ++i) {
    delta_ >> delta[i];
  }

  subsample=iter/1024;
  if(subsample<1) subsample=1;

  dirname = new char[100];

  bool adaptive_ = static_cast<bool>(adaptive);


  bgx(pm, &samples, &conditions, &probes, &genes, &numberOfCategories, samplesets, pm2ps, ps2pm, 
      gamma, delta,
      categories,  
      &iter, &burnin, &s_jmp, &mu_jmp, &sigma_jmp,
      &adaptive_, &batch_size, &optimalAR,
      &subsample, &dirname, &basepath, &seed, &sampleNames[0], filesToCopy);

  delete[] samplesets;
  delete[] pm;
  delete[] categories;
  delete[] dirname;
  delete[] basepath;
  delete[] gamma;
  delete[] delta;
  return 0;
}

// Wrapper to run from R
#ifdef USING_R
extern "C" {
  void mmbgxFromR(char **inputdir, char **basepath) {
    const char *argv[] = {"mmbgx", *inputdir, *basepath};
    main2(3, argv);
  }
}
#endif // USING_R

vector<int> * getIntsByLine(const char *file, int lines) {
  vector<int> *ints = new vector<int>[lines];
  ifstream ifs(file);
  char str[1048576];
  int l=0;
  while(1){
    ifs.getline(str, 1048576);
    if(ifs.eof()) break;
    if(ifs.fail()) {cerr << "In getIntsByLine(): buffer too small. Aborting.\n"; exit(1);}
    char buf[16];
    int i=0;
    while(str[i] != '\0') {
      int j=0;
      while(isdigit(str[i])) {
        buf[j] = str[i];
        j++;
        i++;
      }
      buf[j] = '\0';
      ints[l].push_back(atoi(buf));
      while(!isdigit(str[i]) && str[i] != '\0') i++;
    }
    l++;
  }
  return ints;
}
